/**
 * copyright 2002 2003 Laboratoire d'Informatique Paris 6 (LIP6)
 * 
 * This file is part of ModFact.
 * 
 * ModFact is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * at your option) any later version.
 * 
 * ModFact is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with ModFact; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.modfact.corba.xmiio.exporter;

import org.objectweb.modfact.corba.generator.TagGenerator;
import org.objectweb.modfact.corba.helper.CORBACommon;
import org.objectweb.modfact.corba.logging.Level;
import org.objectweb.modfact.corba.logging.ModFactLogger;
import org.objectweb.modfact.corba.provider.CORBAProvider;
import org.omg.mof.Model.ModelElement;

/**
 * An abstract implementation of an XMI Import.
 * This class provides common functions for all XMI importers,
 * and generic functions for parsing files thanks to the XMIImportUtils inheritance
 */
public abstract class XMIExport extends TagGenerator {

	/** The ORB. */
	protected org.omg.CORBA.ORB _orb;

	/** The CORBA Helper. */
	protected CORBACommon corbaHelper;

	/** The DTD URI. */
	protected String _dtd;

	/** The XML Namespace. */
	protected String _namespace;

	/** The model name. */
	protected String _model;

	/** The Meta-Model name. */
	protected String _metamodel;

	/** The different element exported into XMI with their keys (ref_mof_id / xmi_id). */
	protected java.util.Hashtable _elements = new java.util.Hashtable();

	/** The index for the xmi_id of the elements. */
	protected int _element_index = 1;

	/** The different TypeCode exported into XMI (TypeCode / xmi_id). */
	protected java.util.Hashtable _type_codes = new java.util.Hashtable();

	/** The index for the xmi_id of the TypeCodes. */
	protected int _type_code_index = 1;

	/** The Logger. */
	protected ModFactLogger _logger;

	/**
	 * Default XMI Export Constructor.
	 * @param dtd The DTD File.
	 * @param namespace The XML Namespace.
	 * @param model The model name.
	 * @param metamodel The Meta-Model name.
	 */
	protected XMIExport(String dtd, String namespace, String model, String metamodel) {
		super();
		corbaHelper = new CORBAProvider();
		_dtd = dtd;
		_namespace = namespace;
		_model = model;
		_metamodel = metamodel;
	}

	/**
	 * Set the ORB.
	 * @param orb The new ORB.
	 */
	public void setORB(org.omg.CORBA.ORB orb) {
		_orb = orb;
	}

	/**
	 * Parse the package. The package is retrieve from its name in the Name Service.
	 * @param name The name in the Name Service.
	 * @param printIOR TRUE to display the IOR on the console.
	 */
	public void parseName(String name, boolean printIOR)
		throws
			org.omg.CORBA.ORBPackage.InvalidName,
			org.omg.CosNaming.NamingContextPackage.InvalidName,
			org.omg.CosNaming.NamingContextPackage.NotFound,
			org.omg.CosNaming.NamingContextPackage.CannotProceed,
			org.omg.mof.Reflective.MofError {
		// Get the root naming context
		org.omg.CORBA.Object objRef = _orb.resolve_initial_references("NameService");
		// Use NamingContextExt instead of NamingContext. This is part of the Interoperable naming Service.
		org.omg.CosNaming.NamingContextExt ncRef = org.omg.CosNaming.NamingContextExtHelper.narrow(objRef);
		// Resolve the Object Reference in Naming
		org.omg.CORBA.Object object = ncRef.resolve_str(name);
		parse(object);
		if (printIOR)
			corbaHelper.printIOR(_orb, object, System.out);
	}

	/**
	 * Parse the package. The package is retrieve from an IOR file.
	 * @param iorFile The IOR filename.
	 * @param printIOR TRUE to display the IOR on the console.
	 */
	public void parseIOR(String filename, boolean printIOR) throws org.omg.mof.Reflective.MofError {
		org.omg.CORBA.Object object = corbaHelper.readIOR(_orb, filename);
		parse(object);
		if (printIOR)
			corbaHelper.printIOR(_orb, object, System.out);
	}

	/**
	 * Parse the package.
	 * @param root_package The Root Package.
	 */
	public void parse(org.omg.CORBA.Object root_package) throws org.omg.mof.Reflective.MofError {
		outputln("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		if (_dtd.length() > 0) {
			try {
				java.net.URL url = new java.net.URL(_dtd);
				outputln("<!DOCTYPE XMI SYSTEM \"" + url.toString() + "\">");
			} catch (java.net.MalformedURLException murle) {
				murle.printStackTrace();
			}
		}
		if (_namespace.length() > 0)
			outputln("<XMI verified=\"false\" xmi.version=\"1.1\" xmlns:Model=\"" + _namespace + "\">");
		else
			outputln("<XMI verifier=\"false\" xmi.version=\"1.1\">");
		xmiHeaderTemplate();
		xmiContentTemplate(root_package);
		outputln("</XMI>");
		flushFile();
	}

	/**
	 * Generate the header of the XMI File.
	 */
	public void xmiHeaderTemplate() {
		_logger.log(Level.INFO, "XMI Header generation ...");
		outputln("<XMI.header>");
		outputln("<XMI.model xmi.name=\"" + _model + "\"/>");
		outputln("<XMI.metamodel xmi.name=\"" + _metamodel + "\" xmi.version=\"1.3\"/>");
		outputln("<XMI.documentation>");
		outputln("<XMI.exporter>Generated by the ModFact XMI exporter</XMI.exporter>");
		outputln("</XMI.documentation>");
		outputln("</XMI.header>");
	}

	/**
	 * Generate the content of the XMI File.
	 * @param _root_package The Root Package.
	 */
	public void xmiContentTemplate(org.omg.CORBA.Object _root_package) throws org.omg.mof.Reflective.MofError {
		_logger.log(Level.INFO, "XMI Content generation ...");
		outputln("<XMI.content>");
		rootPackageTemplate(_root_package);
		outputln("</XMI.content>");
	}

	/**
	 * The root package processing.
	 * @param _root_package The Root Package.
	 */
	public abstract void rootPackageTemplate(org.omg.CORBA.Object _root_package) throws org.omg.mof.Reflective.MofError;

	// ==================================================================
	//  Common Functions
	// ==================================================================

	/**
	 * Get the XMI id for an object which is exported to an XMI file.
	 * @param object The object to export.
	 */
	protected String getXMIId(org.omg.mof.Reflective.RefBaseObjectOperations object) {
		String mof_id = object.ref_mof_id();
		if (_elements.containsKey(mof_id))
			return (String) _elements.get(mof_id);
		String key = "a" + _element_index;
		_element_index++;
		_elements.put(mof_id, key);
		return key;
	}

	/**
	 * Get the XMI id for a CORBA TypeCode which is exported to an XMI file.
	 * @param type_code The TypeCode to export.
	 */
	protected String getTypeCodeId(org.omg.CORBA.TypeCode type_code) {
		if (_type_codes.containsKey(type_code))
			return (String) _type_codes.get(type_code);
		String key = "tc" + _type_code_index;
		_type_code_index++;
		_type_codes.put(type_code, key);
		return key;
	}

	/**
	 * Pass from CORBA Type Code to XML.
	 * @param element The XML Element.
	 * @param repositoryID The repository ID.
	 * @return The corresponding TypeCode.
	 */
	public String typeCode2XML(org.omg.CORBA.TypeCode typeCode) throws org.omg.CORBA.BAD_TYPECODE {
		int tc_value = typeCode.kind().value();
		switch (tc_value) {
			case org.omg.CORBA.TCKind._tk_boolean :
				return "XMI.CorbaTcBoolean";
			case org.omg.CORBA.TCKind._tk_octet :
				return "XMI.CorbaTcOctet";
			case org.omg.CORBA.TCKind._tk_char :
				return "XMI.CorbaTcChar";
			case org.omg.CORBA.TCKind._tk_wchar :
				return "XMI.CorbaTcWChar";
			case org.omg.CORBA.TCKind._tk_short :
				return "XMI.CorbaTcShort";
			case org.omg.CORBA.TCKind._tk_ushort :
				return "XMI.CorbaTcUshort";
			case org.omg.CORBA.TCKind._tk_long :
				return "XMI.CorbaTcLong";
			case org.omg.CORBA.TCKind._tk_ulong :
				return "XMI.CorbaTcUlong";
			case org.omg.CORBA.TCKind._tk_longlong :
				return "XMI.CorbaTcLongLong";
			case org.omg.CORBA.TCKind._tk_double :
				return "XMI.CorbaTcDouble";
			case org.omg.CORBA.TCKind._tk_float :
				return "XMI.CorbaTcFloat";
			case org.omg.CORBA.TCKind._tk_string :
				return "XMI.CorbaTcString";
			case org.omg.CORBA.TCKind._tk_TypeCode :
				return "XMI.CorbaTcTypeCode";
			case org.omg.CORBA.TCKind._tk_Principal :
				return "XMI.CorbaTcPrincipal";
			case org.omg.CORBA.TCKind._tk_any :
			default :
				return "XMI.CorbaTcAny";
		}
	}

	/* (non-Javadoc)
	 * @see org.objectweb.modfact.corba.generator.Generator#generate()
	 */
	public void generate() throws Exception {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
	 * @see org.objectweb.modfact.corba.generator.Generator#setInput(org.omg.mof.Model.ModelElement[])
	 */
	public void setInput(ModelElement[] elt) throws Exception {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
	 * @see org.objectweb.modfact.corba.generator.Generator#setLogger(org.objectweb.modfact.corba.logging.ModFactLogger)
	 */
	public void setLogger(ModFactLogger log) throws Exception {
		_logger = log;

	}

}
