/**
 * copyright 2002 2004 Laboratoire d'Informatique Paris 6 (LIP6)
 * 
 * This file is part of ModFact.
 * 
 * ModFact is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * at your option) any later version.
 * 
 * ModFact is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with ModFact; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.modfact.jmi.generator;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;

/**
 * @author Xavier Blanc - Pierre Carpentier
 *
 */
public abstract class PrintGenerator implements Generator {

	/** The output PrintWriter. */
	protected PrintWriter out;
	
	/** The indentation level. */
	protected int level;
	
	/** The tabulation in space characters. */
	protected String TABULATION = "    ";
	
	/** Flag for a new line. */
	protected boolean newLine = true;
	
	/**
	 *  Default Constructor.
	 */
	public PrintGenerator() {
		level = 0;		
	}
	
	/**
	 * Set the output stream.
	 * @param stream The output stream.
	 */
	public void setOutput(OutputStream stream) throws IOException {
		out = new PrintWriter(stream);
	}
	
	/**
	 * Terminate the current line by writing the line separator string.
	 */
	public void outputln()
	{
		out.println();
		newLine = true;
	}
    
    /**
     * Print a String and then terminate the line.
     * @param output The String to output.
     */
	public void outputln(String output)
	{
		output(output);
		outputln();
	}
	
	/**
	 * Print the output to the global PrintWriter (replace out.print(String)).
	 * @param out The string to print.
	 */
	protected void output(String outputString) {
		int index = outputString.indexOf("\n");
		if (index != -1 && index != outputString.length() - 1) {
			// outputString contains at least '\n' which is not at the end
			java.util.StringTokenizer token = new java.util.StringTokenizer(outputString, "\n");
			while (token.hasMoreTokens()) {
				String next = token.nextToken();
				if (!token.hasMoreTokens()) {
					if (outputString.endsWith("\n"))
						out.println(next);
					else
						out.print(next);
				} else {
					out.println(next);
				}
			}
		} else {
			if (outputString.trim().startsWith("}"))
				level--;
			if (newLine) {
				for (int j = 0; j < level; j++)
					out.print(TABULATION);
			}
			if (outputString.trim().startsWith("}")) {
				out.print(outputString);
				if (outputString.trim().endsWith("{"))
					level++;
			} else {
				out.print(outputString);
				for (int i = 0; i < outputString.length(); i++) {
					String current = String.valueOf(outputString.charAt(i));
					if (current.equals("{"))
						level++;
					else if (current.equals("}"))
						level--;
				}
			}
			if (outputString.trim().endsWith("\n"))
				newLine = true;
			else
				newLine = false;
		}
	}
    
    /**
     * Flush the output file.
     */
    public void flushFile() {
        out.flush();
    }
	
	/**
	 * Annotation processing.
	 * @param annotation_ The annotation to print.
	 */
	public void annotationTemplate(String annotation_) {
		if (annotation_.trim().length() > 0) {
			String annotation = "/** \n" + replaceAll(annotation_.trim(), "\\\\n", "\n") + "\n/";
			annotation = replaceAll(annotation, "\\\\s", " ");
			StringBuffer margin = new StringBuffer();
			for (int i = 0; i < level; i++)
				margin.append(TABULATION);
			annotation = replaceAll(annotation, "\n", "\n" + margin.toString() + " *");
			outputln(annotation);
		}
	}
	

	/**
	 * Replace all the occurences of the given string with the given replacement.
	 * @param original The original String to modify.
	 * @param to_replace The substring of the string to search and replace.
	 * @param new_value The replacement String.
	 * @return The modified String.
	 */
	public static String replaceAll(String original, String to_replace, String new_value) {
		StringBuffer newString = new StringBuffer();
		int indexBegin = 0;
		int index = original.indexOf(to_replace, indexBegin);
		while (index != -1) {
			newString.append(original.substring(indexBegin, index));
			newString.append(new_value);
			indexBegin = index + to_replace.length();
			index = original.indexOf(to_replace, indexBegin);
		}
		newString.append(original.substring(indexBegin));
		return newString.toString();
	}
	
}
