package org.objectweb.modfact.qvt.engine;

import java.util.*;
import javax.jmi.reflect.*;

import org.objectweb.modfact.qvt.util.RulePrinter;

class ExpEvaluator {
      
  RuleBinding binding;

  int indent;

  ExpEvaluator(RuleBinding binding) {
    this.binding = binding;
    this.indent = binding.indent+1;
  }
  
  Object eval(RefObject exp) throws Exception {
    try {
      if(Reflection.isOfType(exp, "ParameterExp") ) {
        return parameterExp( exp );

      } else if(Reflection.isOfType(exp, "IfExp") ) {
        return ifExp( exp );

      } else if(Reflection.isOfType(exp, "ModelReferenceExp") ) {
        return modelReferenceExp( exp );

      } else if(Reflection.isOfType(exp, "LiteralExp") ) {
        return literalExp( exp );

      } else if(Reflection.isOfType(exp, "ContextExp") ) {
        return contextExp( exp );
      }
      throw new ExpEvalException("unknown exp: " +RulePrinter.printer.getTypeName(exp), null);
    } catch(Exception e) {
      if(e instanceof ReflectException) {
        if(RuleBinding.verbose) System.err.println( RulePrinter.printer.doIndent(indent+1) +"*** ReflectException :" +e.getMessage());
        if(RuleBinding.verbose) System.err.println( RulePrinter.printer.doIndent(indent+1)
          +"at " +RulePrinter.printer.getTypeName(exp) +": "
          +RulePrinter.printer.printShort(exp)
        );
        Throwable cause = ((ReflectException)e).getCause();
        throw new ExpEvalException("", cause);
      }
      throw e;
    }
  }

  private Object contextExp(RefObject exp) throws Exception {
    Object context = eval((RefObject)exp.refGetValue("context"));
    
    if(Reflection.isOfType(exp, "AbstractOperationCallExp") ) {       
      if(context==null) {
        showNullContextMessage(exp); 
        return null;
      }
      return abstractOperationCallExp( exp, context );

    } else if(Reflection.isOfType(exp, "PropertyCallExp") ) {
      if(context==null) {
        showNullContextMessage(exp); 
        return null;
      }   
      return propertyCallExp( exp, context );

    } else if(Reflection.isOfType(exp, "PropertyAssignExp") ) {
      if(context==null) {
        showNullContextMessage(exp); 
        return null;
      }
      propertyAssignExp( exp, context );
      return null;

    } else if(Reflection.isOfType(exp, "CollectionExp") ) {
       if(context==null) {
          showNullContextMessage(exp); 
          return null;
       }
       return collectionExp( exp, context );

    } else if(Reflection.isOfType(exp, "IsOfTypeExp") ) {
       return isOfTypeExp( exp, context );

    } else if(Reflection.isOfType(exp, "CompareExp") ) {
       return compareExp( exp, context );
    }
    return null;

  }
    
  private void showNullContextMessage(RefObject exp) {
    if(RuleBinding.verbose) {
      System.err.println(
         RulePrinter.printer.doIndent(indent+1)
        +exp.refMetaObject().refGetValue("name")
        +": null context"
      );
    }
  }  

  private Object abstractOperationCallExp (    
    RefObject exp, Object context) throws Exception {
    RefObject[] array = (RefObject[]) ((Collection)exp.refGetValue("args")).toArray(new RefObject[0]);
    List args = new Vector();
    for(int i=0; i<array.length;  i++) {
      args.add(eval(array[i]));
    }
    if( Reflection.isOfType(exp, "RuleCallExp") ) {
      return ruleCallExp( exp, context, args );

    } else {
      return operationCallExp( exp, context, args );
    }
  }

  private Object ruleCallExp(
    RefObject exp, Object context, List argsList) throws Exception {
    if(context==null) {
      return null;
    }
    Object[] argsArray = argsList.toArray();
    RefObject[] params = (RefObject[]) ( (Collection)(
        (RefObject)exp.refGetValue("rule")).refGetValue("params")
    ).toArray(new RefObject[0]);
    Hashtable sharedArgs = new Hashtable();
    for(int i=0; i<params.length; i++ ) {
      sharedArgs.put( (String)params[i].refGetValue("name") ,argsArray[i]);
    }
    RefObject rule = (RefObject)exp.refGetValue("rule");

    if(context instanceof Collection) {
      // iterative call
      Object[] contexts = ((Collection)context).toArray();
      Collection r = new Vector();
      for(int i=0; i<contexts.length; i++) {
        Hashtable args = new Hashtable(sharedArgs);
        args.put("context", contexts[i]);        
        Object aCallResult = new RuleBinding(
            binding.bindingTable
            , rule, binding.src, binding.target, args, indent+1
        ).execute();
        if(aCallResult instanceof Collection) {
          r.addAll((Collection) aCallResult);
        } else if(aCallResult!=null) {
          r.add( aCallResult );
        }
      }
      return r;
    }

    // only one call
    sharedArgs.put("context", context);
    return new RuleBinding(
            binding.bindingTable
            , rule, binding.src, binding.target, sharedArgs, indent+1    
    ).execute();
  }

  private Object operationCallExp(
    RefObject exp, Object context, List args) throws Exception {
    return Reflection.invokeOperation((String)exp.refGetValue("opName"), context ,args);
  }

  private Object parameterExp(RefObject exp) throws Exception {
    return binding.args.get(exp.refGetValue("name")) ;
  }

  private void propertyAssignExp (RefObject exp, Object context) throws Exception {
    Object value = eval((RefObject)exp.refGetValue("value")); //evaluate value
    Reflection.assignProperty((RefObject)context ,(String)exp.refGetValue("propName"), value);
  }

  private Object collectionExp(RefObject exp, Object context) throws Exception {
    Object member = eval((RefObject)exp.refGetValue("member"));
    Collection c = (Collection) context;
    if(member instanceof Collection) {
      Collection members = (Collection) member;
      if (Reflection.isOfType(exp, "AddCollectionExp") ) {
        c.addAll(members);
      } else if (Reflection.isOfType(exp, "RemoveCollectionExp") ) {
        c.removeAll(members);
      }
    } else {
      if (Reflection.isOfType(exp, "AddCollectionExp") ) {
        c.add(member);
      } else if (Reflection.isOfType(exp, "RemoveCollectionExp") ) {
        c.remove(member);
      }
    }
    return c;
  }

  private Object ifExp(RefObject exp) throws Exception {
    Object cond = eval((RefObject)exp.refGetValue("cond"));
    boolean condTestResult;
    if(cond == null ) {
      condTestResult = false;
    } else if ( cond.equals(new Boolean(false)) ) {
      condTestResult = false;
    } else if( cond instanceof Collection
      && ((Collection)cond).isEmpty() ) {
      condTestResult = false;
    } else {
      condTestResult = true;
    }
    if(Reflection.isOfType(exp, "IfStatementExp") ) {
      ifStatementExp( exp ,condTestResult);
      return null;
    } else if(Reflection.isOfType(exp, "IfValueExp") ) {
      return ifValueExp( exp ,condTestResult);
    }
    return null;
  }

  private void ifStatementExp(RefObject exp, boolean condTestResult) throws Exception {
    RefObject[] selectedStatements;
    String thenOrElse;
    if(condTestResult) {
      if(RuleBinding.verbose) System.err.println( RulePrinter.printer.doIndent(indent)
        +"do 'then' {"
      );
      thenOrElse = "thenStatements";
    } else {
      if(RuleBinding.verbose) System.err.println( RulePrinter.printer.doIndent(indent)
        +"do 'else' {"
      );
      thenOrElse = "elseStatements";
    }
    selectedStatements = (RefObject[]) ((Collection)exp.refGetValue(thenOrElse)).toArray(new RefObject[0]);    
    ExpEvaluator evaluator = new ExpEvaluator(binding);    
    for(int i=0; i< selectedStatements.length; i++) {
      if(RuleBinding.verbose) System.err.println( RulePrinter.printer.doIndent(indent+1)
        +(i+1) +": " +RulePrinter.printer.printShort(selectedStatements[i])
      );
      evaluator.eval(selectedStatements[i]);
    }
    if(RuleBinding.verbose) System.err.println( RulePrinter.printer.doIndent(indent)
        +"} //end " +RulePrinter.printer.printShort(exp)
    );    
  }

  private Object ifValueExp(RefObject exp ,boolean condTestResult) throws Exception {
    if(condTestResult) {
      if(RuleBinding.verbose) System.err.println( RulePrinter.printer.doIndent(indent)
        +"do 'then': " +RulePrinter.printer.printShort((RefObject)exp.refGetValue("thenExp"))
      );
      return eval((RefObject)exp.refGetValue("thenExp"));
    } else {
      if(RuleBinding.verbose) System.err.println( RulePrinter.printer.doIndent(indent)
        +"do 'else': " +RulePrinter.printer.printShort((RefObject)exp.refGetValue("elseExp"))
      );
      return eval((RefObject)exp.refGetValue("elseExp"));
    }
  }

  private Object modelReferenceExp(RefObject exp) throws Exception {
    return Reflection.allOfType(binding.src, (String)exp.refGetValue("conceptName")) ;
  }


  private Object propertyCallExp(RefObject exp, Object context) throws Exception {
    return Reflection.getPropertyValue(
      (RefObject)context, (String)exp.refGetValue("propName")
    );
  }

  private Object compareExp(RefObject exp, Object context) throws Exception {
    Object val = eval((RefObject)exp.refGetValue("compareVal"));
    if(context==null) return new Boolean(val==null);
    return new Boolean(context.equals(val));
  }

  private Object isOfTypeExp(RefObject exp, Object context) throws Exception {
    if(context==null) 
        return new Boolean(false);
    return new Boolean(
        Reflection.isOfType((RefObject)context, (String)exp.refGetValue("type"))
    );
  }

  private Object literalExp(RefObject exp) throws Exception {

	  if(Reflection.isOfType(exp, "DoubleLiteral") ) {
		  return (Double) exp.refGetValue("val");

	  } else if(Reflection.isOfType(exp, "IntegerLiteral") ) {
		  return (Integer) exp.refGetValue("val");

	  } else if(Reflection.isOfType(exp, "BooleanLiteral") ) {
		  return (Boolean) exp.refGetValue("val");

	  } else if(Reflection.isOfType(exp, "StringLiteral") ) {
		  return (String) exp.refGetValue("val");

	  } else if(Reflection.isOfType(exp, "EnumLiteral") ) {
		  return Reflection.getEnumValue( 
                    srcOrTarget((Boolean) exp.refGetValue("isFromMmSrc"))
                  , (String) exp.refGetValue("type")
                  , (String) exp.refGetValue("val") );

	  }
	  return null;
  }
  
  private RefPackage srcOrTarget(Boolean isFromMmSrc) {
    return (isFromMmSrc.equals(Boolean.TRUE))? binding.src: binding.target;
  }
  
}



